/**
 * 
 */
package isen.java.library;

import isen.java2.library.exceptions.ItemAlreadyBorrowedException;
import isen.java2.library.exceptions.ItemNotFoundException;
import isen.java2.library.model.Book;
import isen.java2.library.model.CulturalItem;
import isen.java2.library.model.Movie;
import isen.java2.library.model.Genre;

/**
 * This is our main library class, which will hold all the logic of
 * CulturalItems handling
 * 
 * @author Philippe Duval
 */
public class Library {

	// Private field for Movies handling. note that we make the choice to use a
	// list, which allows us to have duplicates. We use generics to specialize
	// the objects used in the list.
	// you could use another storage structure, as long as it is sortable, as
	// it's required for the exercise
	// Note that we use an abstract type, the implementation is up to the caller.
	private List<Movie> listOfFilms;

	// Private field for books handling.
	private List<Book> listOfBooks;

	/**
	 * Simple parameterized constructor
	 * 
	 * @param listOfMovies
	 *                    the list of Movies available in the library
	 * @param listOfBooks
	 *                    the list of books available in the library
	 */
	public Library(List<Movie> listOfMovies, List<Book> listOfBooks) {
		this.listOfMovies = listOfMovies;
		this.listOfBooks = listOfBooks;
	}

	/**
	 * Adds a Movie to the library
	 * 
	 * @param Movie
	 */
	public void add(Movie movie) {
		this.listOfMovies.add(aMovie);
	}

	/**
	 * Adds a book to the library
	 * 
	 * @param book
	 */
	public void add(Book book) {
		book.add(this.listOfBook);
	}

	/**
	 * Allows a given user to borrow a Movie
	 * 
	 * @param MovieTitle
	 * @param borrower
	 * @return
	 * @throws ItemNotFoundException
	 *                                      if the title of the Movie does not match
	 *                                      an existing item in
	 *                                      the library
	 * @throws ItemAlreadyBorrowedException
	 *                                      if the Movie is already borrowed
	 */
	public Movie borrowMovie(String movieTitle, String borrower)
			throws ItemNotFoundException, ItemAlreadyBorrowedException {
		return borrowItem(movieTitle, borrower, this.listOfMovies);
	}

	/**
	 * 
	 * @param bookTitle
	 * @param borrower
	 * @return
	 * @throws ItemNotFoundException
	 *                                      if the title of the Movie does not match
	 *                                      an existing item in
	 *                                      the library
	 * @throws ItemAlreadyBorrowedException
	 *                                      if the Movie is already borrowed
	 */
	public Book borrowBook(String bookTitle, String borrower)
			throws ItemNotFoundException, ItemAlreadyBorrowedException {
		return borrowItem(bookTitle, borrower, this.listOfBooks);
	}

	/**
	 * As the borrowBook and borrowMovie implement practically the same logic, we
	 * factorize their code in a private method that does all the logic using
	 * generics. we will use a generic T extending CulturalItem to parameterize
	 * our class. this generic will be inferred at compilation time depending on
	 * the calling method's signature
	 * 
	 * @param itemTitle
	 * @param borrower
	 * @return
	 * @throws ItemAlreadyBorrowedException
	 * @throws ItemNotFoundException
	 */
	private <T extends CulturalItem> T borrowItem(String itemTitle, String borrower, List<T> listOfItems)
			throws ItemNotFoundException {
		T item = this.searchByTitle(itemTitle, listOfItems);
		if (item.isBorrowed()) {
			throw new ItemNotFoundException();
		} else {
			item.borrow(borrower);
			return item;
		}
	}

	/**
	 * Method listing what was borrowed by a certain user
	 * 
	 * @param borrower
	 */
	public void printBorrowedItems(String borrower) {
		this.printSortedList(this.filterByBorrower(this.listOfBooks, borrower), "Borrowed Books by " + borrower);
		this.printSortedList(this.filterByBorrower(this.listOfMovies, borrower), "Borrowed Movies by " + borrower);
	}

	/**
	 * Method allowing to rate a Movie
	 * 
	 * @param MovieTitle
	 * @param person
	 * @param rating
	 * @throws ItemNotFoundException
	 */
	public void rateMovie(String movieTitle, String person, Integer rating) throws ItemNotFoundException {
		Movie movie = this.searchByTitle(movieTitle, this.listOfMovies);
		movie.addRating(person, rating);
	}

	/**
	 * Method allowing to rate a book
	 * 
	 * @param bookTitle
	 * @param person
	 * @param rating
	 * @throws ItemNotFoundException
	 */
	public void rateBook(String bookTitle, String person, Integer rating) throws ItemNotFoundException {
		Book book = this.searchByTitle(bookTitle, this.listOfBooks);
		book.addRating(person, rating);
	}

	/**
	 * Method returning the top ten items according to all ratings in the
	 * library
	 */
	public void printTopTenItems() {
		List<CulturalItem> top = getTopTenItems();
		// print list (TRAP: you cannot use the print method, as it already
		// makes sorting for you)
		// here we use a ternary operator to extract the sublist, in order to
		// avoid a indexOutOfBoundException
		this.printList(top.size() > 10 ? top.subList(0, 10) : top, "Our Best Of");
	}

	/**
	 * Method returning the top ten items
	 * Beware that it can return less than ten articles!
	 * 
	 * @return a @List of culturalItems
	 */
	public List<CulturalItem> getTopTenItems() {
		// Create a list with all items, by aggregating the two lists.
		List<CulturalItem> top = new LinkedList<>(this.listOfBooks);
		top.addAll(this.listOfMovies);
		// sort by averageRating, using an inner comparator class
		Collections.sort(top, new Comparator<CulturalItem>() {

			@Override
			public int compare(CulturalItem o1, CulturalItem o2) {
				// as averageRatting returns a double, the implementation is
				// straightforward
				return o1.getAverageRating().compareTo(o2.getAverageRating());
			}

		});
		return top.subList(0, 10);
	}

	/**
	 * Here again, we use a private method using generics to factorize our code.
	 * Note that in the case we want to handle another type of CulturalItem, all
	 * these private generic methods stay unchanged
	 * 
	 * @param title
	 * @return
	 */
	private <T extends CulturalItem> T searchByTitle(String title, List<T> items) throws ItemNotFoundException {
		for (T item : items) {
			if (item.getTitle().equals(title)) {
				return item;
			}
		}
		// note: if we find the item in the previous loop, we will never reach
		// this code, so the exception wont be thrown.
		// if we don't find the book, we exit the loop, and throw the exception.
		// there is no need for a local variable here.
		throw new ItemNotFoundException();
	}

	/**
	 * Method printing the two lists of our library, sorted by alphabetical
	 * order
	 */
	public void printCatalogue() {
		this.printSortedList(this.listOfBooks, "Our Books");
		this.printSortedList(this.listOfMovies, "Our Movies");
	}

	/**
	 * Method returning a list of our library
	 */
	public List<CulturalItem> getCatalogue() {
		List<CulturalItem> items = new LinkedList<>();
		items.addAll(listOfBooks);
		items.addAll(listOfMovies);
		Collections.sort(items);
		return items;
	}

	/**
	 * Method searching items by genre and printing the result in alphabetical
	 * order
	 * 
	 * @param genre
	 */
	public void printByGenre(Genre genre) {
		this.printSortedList(this.filterByGenre(genre, this.listOfBooks), "Books of " + genre);
		this.printSortedList(this.filterByGenre(genre, this.listOfMovies), "Movies of " + genre);

	}

	/**
	 * Method searching items by genre and printing the result in alphabetical
	 * order
	 * 
	 * @param genre
	 */
	public List<CulturalItem> getByGenre(Genre genre) {
		List<CulturalItem> items = new LinkedList<>();
		items.addAll(this.filterByGenre(genre, this.listOfBooks));
		items.addAll(this.filterByGenre(genre, this.listOfMovies));
		return items;
	}


	/**
	 * And yet another private method using generics, this time to filter our
	 * collections by genre
	 * 
	 * @param genre
	 * @param items
	 * @return
	 */
	private <T extends CulturalItem> List<T> filterByGenre(Genre genre, List<T> items) {

		// Here we have to use a local variable : remember that the items list
		// though passed by value, still reference our object's list. filtering
		// it directly without passing by a local variable would be permanent !
		List<T> filteredList = new LinkedList<>(items);

		// we use an inner class that is a predicate, which allows us to filter
		// the collection without having to call an iterator
		filteredList.removeIf(new Predicate<CulturalItem>() {

			@Override
			public boolean test(CulturalItem t) {
				return !t.getGenres().contains(genre);
			}
		});
		return filteredList;
	}

	/**
	 * Method searching items by genre and printing the result in alphabetical
	 * order
	 * 
	 * @param genre
	 */
	public List<CulturalItem> getByBorrower(String borrower) {
		List<CulturalItem> items = new LinkedList<>();
		items.addAll(this.filterByBorrower(this.listOfBooks, borrower));
		items.addAll(this.filterByBorrower(this.listOfMovies, borrower));
		return items;
	}

	/**
	 * Method allowing us to filter by borrower a list of items. Again, we use
	 * generics.
	 * 
	 * @param items
	 * @param borrower
	 * @param genre
	 * @return
	 */
	private <T extends CulturalItem> List<T> filterByBorrower(List<T> items, String borrower) {
		// again we filter a local copy of our list, using a predicate.
		List<T> filteredList = new LinkedList<>(items);

		filteredList.removeIf(new Predicate<CulturalItem>() {

			@Override
			public boolean test(CulturalItem t) {
				// filters the list by excluding the items not borrowed, or
				// borrowed by another person
				return (!t.isBorrowed()) || (t.isBorrowed() && !t.getBorrower().equals(borrower));
			}
		});
		return filteredList;
	}

	/**
	 * You bet it ! private method using generics, to print a list of items
	 * 
	 * @param items
	 * @param type
	 */
	private <T extends CulturalItem> void printSortedList(List<T> items, String type) {
		// Handy method allowing us to sort a list by it's natural order (see
		// the cultural item class for further explanation)
		Collections.sort(items);
		this.printList(items, type);
	}

	/**
	 * @param items
	 */
	private <T extends CulturalItem> void printList(List<T> items, String type) {
		System.out.println("======= " + type + " =============");
		// this way to iterate is not the shortest way to print our list, but
		// demonstrates another use of inner classes, this time with a consumer
		items.forEach(new Consumer<T>() {

			@Override
			public void accept(T it) {
				it.print();

			}
		});
	}
}
